#include <TMB.hpp>
template<class Type>
Type objective_function<Type>::operator() ()
{
  using namespace R_inla;
  using namespace density;
  using namespace Eigen;
  
  // Input Data
  
  DATA_INTEGER(bigN); // number of pixels in image plane
  DATA_MATRIX(evi_covariate_matrix); // [dim: bigN x 12]
  DATA_MATRIX(LSTday_covariate_matrix); // [dim: bigN x 12]
  DATA_MATRIX(LSTdiff_covariate_matrix); // [dim: bigN x 12]
  DATA_MATRIX(TCB_covariate_matrix); // [dim: bigN x 12]
  DATA_MATRIX(TCW_covariate_matrix); // [dim: bigN x 12]

  DATA_STRUCT(spde,spde_t); // INLA SPDE object (components of precision matrix)
  DATA_SPARSE_MATRIX(A); // INLA SPDE projection matrix: mesh to pixels [dim: bigN x nMesh]
  
  DATA_INTEGER(nHFs); // number of (aggregated) health facilities
  DATA_VECTOR(population); // [length: bigN]
  DATA_MATRIX(HFcases_matrix); // [dim: nHFs x 12]  
  DATA_VECTOR(log_masses); // [length: nHFs]
  DATA_VECTOR(baseline_surface); // [length: bigN]

  DATA_INTEGER(Nunwrapped); // number of HF-labelled pixels with non-zero catchment population
  DATA_VECTOR(invdists); // [length: Nunwrapped]
  DATA_IVECTOR(hf_ids); // [length: Nunwrapped]
  DATA_IVECTOR(pixel_ids); // [length: Nunwrapped]
  
  DATA_VECTOR(treatment); // [length: bigN]
  
  DATA_MATRIX(intermonth_distmat); // [dim: 12 x 12]
  
  // Parameters
  
  PARAMETER(intercept);
  PARAMETER(log_overdispersion_field_sd);
  PARAMETER(log_shrinkage_temporal_slopes);
  PARAMETER(evi_slope_lag0);
  PARAMETER(evi_slope_lag1);
  PARAMETER(evi_slope_lag2);
  PARAMETER(LSTday_slope_lag0);
  PARAMETER(LSTday_slope_lag1);
  PARAMETER(LSTday_slope_lag2);
  PARAMETER(LSTdiff_slope_lag0);
  PARAMETER(LSTdiff_slope_lag1);
  PARAMETER(LSTdiff_slope_lag2);
  PARAMETER(TCB_slope_lag0);
  PARAMETER(TCB_slope_lag1);
  PARAMETER(TCB_slope_lag2);
  PARAMETER(TCW_slope_lag0);
  PARAMETER(TCW_slope_lag1);
  PARAMETER(TCW_slope_lag2);
  PARAMETER(log_range_seasonal);
  PARAMETER(log_sd_seasonal);
  PARAMETER(log_range_seasonaltime);
  PARAMETER(log_sd_seasonaltime);
  PARAMETER_ARRAY(field_seasonal); // [dim: nMesh x 12]

  // Parameter Transforms

  Type overdispersion_field_sd = exp(log_overdispersion_field_sd);
  Type shrinkage_temporal_slopes = 1.0;//exp(log_shrinkage_temporal_slopes);
  Type range_seasonal = exp(log_range_seasonal);
  Type kappa_seasonal = 2.8284/range_seasonal;
  Type sd_seasonal = exp(log_sd_seasonal);
  Type range_seasonaltime = exp(log_range_seasonaltime);
  Type var_seasonaltime = exp(log_sd_seasonaltime*2.0);

  vector<Type> masses = exp(log_masses);

  // Priors
  
  Type nll = 0.0;
  
  nll -= dnorm(log_overdispersion_field_sd,Type(-2.0),Type(1.0),true);
  
  nll -= dnorm(log_shrinkage_temporal_slopes,Type(0.0),Type(1.0),true);
  nll -= dnorm(evi_slope_lag0,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(evi_slope_lag1,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(evi_slope_lag2,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(LSTday_slope_lag0,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(LSTday_slope_lag1,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(LSTday_slope_lag2,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(LSTdiff_slope_lag0,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(LSTdiff_slope_lag1,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(LSTdiff_slope_lag2,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(TCB_slope_lag0,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(TCB_slope_lag1,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(TCB_slope_lag2,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(TCW_slope_lag0,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(TCW_slope_lag1,Type(0.0),shrinkage_temporal_slopes,true);
  nll -= dnorm(TCW_slope_lag2,Type(0.0),shrinkage_temporal_slopes,true);
  
  nll -= dnorm(log_sd_seasonal,Type(-1),Type(1.0),true);
  nll -= dnorm(log_range_seasonal,Type(1),Type(1.0),true);
  nll -= dnorm(log_sd_seasonaltime,Type(-1),Type(1.0),true);
  nll -= dnorm(log_range_seasonaltime,Type(1),Type(1.0),true);
  SparseMatrix<Type> Q_seasonal = Q_spde(spde,kappa_seasonal);
  matrix<Type> tempcov(12,12);
  for (int i=0; i<12; i++) {
    for (int j=0; j<12; j++) {
      tempcov(i,j) = var_seasonaltime*exp(-intermonth_distmat(i,j)/range_seasonaltime);
    }
  }
  for (int i=0; i<12; i++) {
    tempcov(i,i) += 0.0001;
  }
  nll += SEPARABLE(MVNORM(tempcov),SCALE(GMRF(Q_seasonal),sd_seasonal))(field_seasonal);
  
  // Algebra: Construct baseline field 
   
  matrix<Type> seasonal_field(bigN,12);
  seasonal_field = A*field_seasonal.matrix();

  matrix<Type> predicted_surface_malaria(bigN,12);
  for (int i=0; i<bigN; i++) {
    for (int j=0; j<12; j++) {
      predicted_surface_malaria(i,j) = intercept + evi_covariate_matrix(i,j)*evi_slope_lag0 + evi_covariate_matrix(i,(j+11) % 12)*evi_slope_lag1 + evi_covariate_matrix(i,(j+10) % 12)*evi_slope_lag2;
      predicted_surface_malaria(i,j) += LSTday_covariate_matrix(i,j)*LSTday_slope_lag0 + LSTday_covariate_matrix(i,(j+11) % 12)*LSTday_slope_lag1 + LSTday_covariate_matrix(i,(j+10) % 12)*LSTday_slope_lag2;
      predicted_surface_malaria(i,j) += LSTdiff_covariate_matrix(i,j)*LSTdiff_slope_lag0 + LSTdiff_covariate_matrix(i,(j+11) % 12)*LSTdiff_slope_lag1 + LSTdiff_covariate_matrix(i,(j+10) % 12)*LSTdiff_slope_lag2;
      predicted_surface_malaria(i,j) += TCB_covariate_matrix(i,j)*TCB_slope_lag0 + TCB_covariate_matrix(i,(j+11) % 12)*TCB_slope_lag1 + TCB_covariate_matrix(i,(j+10) % 12)*TCB_slope_lag2;
      predicted_surface_malaria(i,j) += TCW_covariate_matrix(i,j)*TCW_slope_lag0 + TCW_covariate_matrix(i,(j+11) % 12)*TCW_slope_lag1 + TCW_covariate_matrix(i,(j+10) % 12)*TCW_slope_lag2;
      predicted_surface_malaria(i,j) += baseline_surface[i];
      predicted_surface_malaria(i,j) += seasonal_field(i,j);
    }
  }
  
  matrix<Type> predicted_surface_malaria_effrate(bigN,12);
  for (int i=0; i<bigN; i++) {
    for (int j=0; j<12; j++) {
      predicted_surface_malaria_effrate(i,j) = exp(predicted_surface_malaria(i,j))*population[i]*treatment[i];
    }
  }
  
  // Algebra: Catchments
  
  vector<Type> catchment_weights(Nunwrapped);
  vector<Type> catchment_totals(bigN);
  for (int i=0; i<Nunwrapped; i++) {
    catchment_weights[i] = invdists[i]*masses[hf_ids[i]-1];
  }
  for (int i=0; i<bigN; i++) {catchment_totals[i] = 0.0;}
  for (int i=0; i<Nunwrapped; i++) {
    catchment_totals[pixel_ids[i]-1] += catchment_weights[i];
  }
  for (int i=0; i<Nunwrapped; i++) {
    catchment_weights[i] = catchment_weights[i]/catchment_totals[pixel_ids[i]-1];
  }
  std::vector< Eigen::Triplet<Type> > catchment_vals;
  catchment_vals.reserve(Nunwrapped);
  for (int i=0; i<Nunwrapped; i++) {
    catchment_vals.push_back(Eigen::Triplet<Type>(hf_ids[i]-1,pixel_ids[i]-1,catchment_weights[i]));
  }
  SparseMatrix<Type> catchments(nHFs,bigN);
  catchments.setFromTriplets(catchment_vals.begin(),catchment_vals.end());
  
  // Algebra: Expected Cases

  matrix<Type> predicted_cases(nHFs,12);
  predicted_cases = catchments*predicted_surface_malaria_effrate;

  // Likelihood

  for (int i=0; i<nHFs; i++) {
    for (int j=0; j<12; j++) {
      nll -= dnbinom2(HFcases_matrix(i,j),predicted_cases(i,j),predicted_cases(i,j)*(1.0+overdispersion_field_sd),true);
    }
  }
  
  // Reporting

  REPORT(predicted_surface_malaria);
  REPORT(predicted_cases);
  REPORT(seasonal_field);

  return nll;
}
